{******************************************************************************}
{                                                                              }
{           ReportBuilder Data Access Developement Environment (DADE)          }
{                                                                              }
{             Copyright (c) 1996, 2002 Digital Metaphors Corporation           }
{                                                                              }
{             Copyright (c) 2002 Liodden Data AS                               }
{                                                                              }
{******************************************************************************}

unit daNDB;

interface

{$I ppIfDef.pas}

uses Classes, SysUtils, Forms, ExtCtrls, DB, Contnrs,
     ppClass, ppComm, ppDBPipe, ppDB, ppClasUt, ppTypes,
     daDB, daQueryDataView, daDataView, daPreviewDataDlg,
     NdbAsa, NdbAsaLg, NdbAsaDS, NdbApp;

type

  {NativeDB for SQL Anywhere (NDB) DataView Classes:

     1.  NDB TDataSet descendants
           - TDataSets that can be children of a DataView.
           - Override the HasParent method of TComponent to return True
           - Must be registerd with the Delphi IDE using the RegisterNoIcon procedure

       a. TdaChildNDBQuery        - TAsaDataSet descendant that can be a child of a DataView
       b. TdaChildNDBTable        - TAsaDataSet descendant that can be a child of a DataView
       c. TdaChildNDBStoredProc   - TAsaStoredProc descendant that can be a child of a DataView

     2.  TdaNDBSession
           - descendant of TppSession
           - implements GetDatabaseNames, GetTableNames, etc.

     3.  TdaNDBDataSet
          - descendant of TppDataSet
          - implements GetFieldNames for SQL

     4.  TdaNDBQueryDataView
          - descendant of TppQueryDataView
          - uses the above classes to create the required
            Query -> DataSource -> Pipeline -> Report connection
          - uses the TdaSQL object built by the QueryWizard to assign
            SQL to the TNDBQuery etc.
      }

  { TdaChildNDBQuery }
  TdaChildNDBQuery = class(TAsaDataset)
    public
      function HasParent: Boolean; override;
    end;

  { TdaChildNDBTable }
  TdaChildNDBTable = class(TAsaDataset)
    public
      function HasParent: Boolean; override;
    end;

  { TdaChildNDBStoredProc }
  TdaChildNDBStoredProc = class(TAsaStoredProc)
    public
      function HasParent: Boolean; override;
  end;

  { TdaNDBSession }
  TdaNDBSession = class(TdaSession)
    private
      procedure AddDatabase(aDatabase: TComponent);
    protected
      procedure SetDataOwner(aDataOwner: TComponent); override;
    public
      class function ClassDescription: String; override;
      class function DataSetClass: TdaDataSetClass; override;
      class function DatabaseClass: TComponentClass; override;
      procedure GetDatabaseNames(aList: TStrings); override;
      function  GetDatabaseType(const aDatabaseName: String): TppDatabaseType; override;
      procedure GetTableNames(const aDatabaseName: String; aList: TStrings); override;
      function  ValidDatabaseTypes: TppDatabaseTypes; override;
  end;

  { TdaNDBDataSet }
  TdaNDBDataSet = class(TdaDataSet)
    private
      FQuery: TAsaDataset;
      function GetQuery: TAsaDataset;
    protected
      procedure BuildFieldList; override;
      function  GetActive: Boolean; override;
      procedure SetActive(Value: Boolean); override;
      procedure SetDatabaseName(const aDatabaseName: String); override;
      procedure SetDataName(const aDataName: String); override;
      property Query: TAsaDataset read GetQuery;
    public
      constructor Create(aOwner: TComponent); override;
      destructor Destroy; override;
      class function ClassDescription: String; override;
      procedure GetFieldNamesForSQL(aList: TStrings; aSQL: TStrings); override;
      procedure GetFieldsForSQL(aList: TList; aSQL: TStrings); override;
  end;

  { TdaNDBQueryDataView }
  TdaNDBQueryDataView = class(TdaQueryDataView)
    private
      FDataSource: TppChildDataSource;
      FQuery: TdaChildNDBQuery;
    protected
      procedure SQLChanged; override;
    public
      constructor Create(aOwner: TComponent); override;
      destructor Destroy; override;
      class function PreviewFormClass: TFormClass; override;
      class function SessionClass: TClass; override;
      procedure Init; override;
      procedure ConnectPipelinesToData; override;
    published
      property DataSource: TppChildDataSource read FDataSource;
  end;

  {global functions to access default NDB connection}
  function daGetDefaultNDBConnection: TAsaSession;

  {utility routines}
  procedure daGetNDBConnectionNames(aList: TStrings);
  function daGetNDBConnectionForName(aDatabaseName: String): TAsaSession;
  function daNDBConnectToDatabase(aConnection: TAsaSession): Boolean;

  function daGetNDBConnectionList: TppComponentList;

  {Delphi design time registration}
  procedure Register;

implementation

const
  cDefaultConnection = 'DefaultNDBConnection';

var
  FNDBConnection: TAsaSession;
  FNDBLogin: TAsaLogin;
  FNDBConnectionList: TppComponentList;

{******************************************************************************
 *
 ** R E G I S T E R
 *
{******************************************************************************}

procedure Register;
begin
  {NDB DataAccess Components}
  RegisterNoIcon([TdaChildNDBQuery, TdaChildNDBTable, TdaChildNDBStoredProc]);
  {NDB DataViews}
  RegisterNoIcon([TdaNDBQueryDataView]);
end;

{******************************************************************************
 *
 ** C H I L D  N A T I V E D B  D A T A   A C C E S S   C O M P O N  E N T S
 *
{******************************************************************************}

{------------------------------------------------------------------------------}
{ TdaChildNDBQuery.HasParent }

function TdaChildNDBQuery.HasParent: Boolean;
begin
  Result := True;
end; {function, HasParent}

{------------------------------------------------------------------------------}
{ TdaChildNDBTable.HasParent }

function TdaChildNDBTable.HasParent: Boolean;
begin
  Result := True;
end; {function, HasParent}

{------------------------------------------------------------------------------}
{ TdaChildNDBStoredProc.HasParent }

function TdaChildNDBStoredProc.HasParent: Boolean;
begin
  Result := True;
end; {function, HasParent}

{******************************************************************************
 *
 ** N A T I V E D B   S E S S I O N
 *
{******************************************************************************}

{------------------------------------------------------------------------------}
{ TdaNDBSession.ClassDescription }

class function TdaNDBSession.ClassDescription: String;
begin
  Result := 'NDBSession';
end; {class function, ClassDescription}

{------------------------------------------------------------------------------}
{ TdaNDBSession.DataSetClass }

class function TdaNDBSession.DataSetClass: TdaDataSetClass;
begin
  Result := TdaNDBDataSet;
end; {class function, DataSetClass}

{------------------------------------------------------------------------------}
{ TdaNDBSession.DatabaseClass }

class function TdaNDBSession.DatabaseClass: TComponentClass;
begin
  Result := TAsaSession;
end;

{------------------------------------------------------------------------------}
{ TdaNDBSession.GetDatabaseType }

function TdaNDBSession.GetDatabaseType(const aDatabaseName: String): TppDatabaseType;
begin
  Result := dtSybaseASA;
end; {procedure, GetDatabaseType}

{------------------------------------------------------------------------------}
{ TdaNDBSession.GetTableNames }

procedure TdaNDBSession.GetTableNames(const aDatabaseName: String; aList: TStrings);
var
  lDatabase: TAsaSession;
  lsTableName: String;
  lQuery: TAsaDataset;

begin
  aList.Clear;

  {get the database}
  lDatabase := daGetNDBConnectionForName(aDatabaseName);

  {connection must be active to get table names}
  if not daNDBConnectToDatabase(lDatabase) then Exit;

  {get list of table names from a table object}
  lQuery := TAsaDataset.Create(nil);
  try
    lQuery.Session := daGetNDBConnectionForName(aDatabaseName);

    lQuery.SQL.Text :=   'select creator, tname ' +
                         'from syscatalog ' +
                         'where (tabletype = ''TABLE'' ' +
                         'or tabletype = ''VIEW'') ' +
                         'and creator not in (''SYS'', ''dbo'', ''rs_systabgroup'')';

    lQuery.Open;

    if lQuery.Active then
      begin

        lQuery.First;

        while not lQuery.EOF do
          begin
            lsTableName := lQuery.Fields[0].AsString + '.' + lQuery.Fields[1].AsString;
            aList.Add(lsTableName);

            lQuery.Next;

          end;
      end;

  finally
    lQuery.Free;
  end;

end; {procedure, GetTableNames}

{------------------------------------------------------------------------------}
{ TdaNDBSession.AddDatabase }

procedure TdaNDBSession.AddDatabase(aDatabase: TComponent);
begin

  if daGetNDBConnectionList.IndexOf(aDatabase) < 0 then
    FNDBConnectionList.Add(aDatabase);

end; {procedure, AddDatabase}

{------------------------------------------------------------------------------}
{ TdaNDBSession.GetDatabaseNames }

procedure TdaNDBSession.GetDatabaseNames(aList: TStrings);
var
  liIndex: Integer;

begin
  {call utility routine to get list of database names}
  daGetNDBConnectionNames(aList);

  daGetDatabaseObjectsFromOwner(TdaSessionClass(Self.ClassType), aList, DataOwner);

  for liIndex := 0 to aList.Count-1 do
    if aList.Objects[liIndex] <> nil then
      AddDatabase(TComponent(aList.Objects[liIndex]));

end; {procedure, GetDatabaseNames}

{------------------------------------------------------------------------------}
{ TdaNDBSession.SetDataOwner }

procedure TdaNDBSession.SetDataOwner(aDataOwner: TComponent);
var
  lList: TStringList;
begin

  inherited SetDataOwner(aDataOwner);

  lList := TStringList.Create;
  try
    GetDatabaseNames(lList);
  finally
    lList.Free;
  end;
end; {procedure, SetDataOwner}

{------------------------------------------------------------------------------}
{ TdaNDBSession.ValidDatabaseTypes }

function TdaNDBSession.ValidDatabaseTypes: TppDatabaseTypes;
begin
  Result := [dtSybaseASA];

end; {function, ValidDatabaseTypes}

{******************************************************************************
 *
 ** N A T I V E D B   D A T A S E T
 *
{******************************************************************************}

{------------------------------------------------------------------------------}
{ TdaNDBDataSet.Create }

constructor TdaNDBDataSet.Create(aOwner: TComponent);
begin

  inherited Create(aOwner);

  FQuery := nil;

end; {constructor, Create}

{------------------------------------------------------------------------------}
{ TdaNDBDataSet.Destroy }

destructor TdaNDBDataSet.Destroy;
begin

  FQuery.Free;

  inherited Destroy;

end; {destructor, Destroy}

{------------------------------------------------------------------------------}
{ TdaNDBDataSet.ClassDescription }

class function TdaNDBDataSet.ClassDescription: String;
begin
  Result := 'NDBDataSet';
end; {class function, ClassDescription}

{------------------------------------------------------------------------------}
{ TdaNDBDataSet.GetActive }

function TdaNDBDataSet.GetActive: Boolean;
begin
  Result := GetQuery.Active
end; {function, GetActive}

{------------------------------------------------------------------------------}
{ TdaNDBDataSet.SetActive }

procedure TdaNDBDataSet.SetActive(Value: Boolean);
begin
  GetQuery.Active := Value;
end; {procedure, SetActive}

{------------------------------------------------------------------------------}
{ TdaNDBDataSet.GetQuery }

function TdaNDBDataSet.GetQuery: TAsaDataset;
begin

  {create NDB DataSet, if needed}
  if (FQuery = nil) then
    FQuery := TAsaDataset.Create(Self);

  Result := FQuery;

end; {procedure, GetQuery}

{------------------------------------------------------------------------------}
{ TdaNDBDataSet.SetDatabaseName }

procedure TdaNDBDataSet.SetDatabaseName(const aDatabaseName: String);
begin

  inherited SetDatabaseName(aDatabaseName);

  {table cannot be active to set database property}
  if GetQuery.Active then
    FQuery.Active := False;

  {get NDB Connection for name}
  FQuery.Session := daGetNDBConnectionForName(aDatabaseName);

end; {procedure, SetDatabaseName}

{------------------------------------------------------------------------------}
{ TdaNDBDataSet.SetDataName }

procedure TdaNDBDataSet.SetDataName(const aDataName: String);
begin

  inherited SetDataName(aDataName);

  {dataset cannot be active to set data name}
  if GetQuery.Active then
    FQuery.Active := False;

  {construct an SQL statment that returns an empty result set,
   this is used to get the field information }
  FQuery.SQL.Text := 'SELECT * FROM ' + aDataName + ' WHERE ''c'' <> ''c'' ';

end; {procedure, SetDataName}

{------------------------------------------------------------------------------}
{ TdaNDBDataSet.BuildFieldList }

procedure TdaNDBDataSet.BuildFieldList;
var
  liIndex: Integer;
  lQueryField: TField;
  lField: TppField;
begin

  inherited BuildFieldList;

  {set table to active}
  if not(GetQuery.Active) then
    FQuery.Active := True;

  {create TppField objects for each field in the table}
  for liIndex := 0 to FQuery.FieldCount - 1 do
    begin
      lQueryField := FQuery.Fields[liIndex];

      lField := TppField.Create(nil);

      lField.TableName := DataName;
      lField.FieldName := lQueryField.FieldName;
      lField.FieldAlias := lQueryField.DisplayLabel;
      lField.FieldLength := lQueryField.Size;
      lField.DataType := ppConvertFieldType(lQueryField.DataType);
      lField.DisplayWidth := lQueryField.DisplayWidth;

      AddField(lField);
    end;

end; {function, BuildFieldList}

{------------------------------------------------------------------------------}
{ TdaNDBDataSet.GetFieldNamesForSQL }

procedure TdaNDBDataSet.GetFieldNamesForSQL(aList: TStrings; aSQL: TStrings);
var
  lQuery: TAsaDataset;
begin
  aList.Clear;

  {create a temporary query}
  lQuery := TAsaDataset.Create(Self);
  try

    {set the database and SQL properties}
    lQuery.Session := daGetNDBConnectionForName(DatabaseName);
    lQuery.SQL := aSQL;

    {get the field names}
    lQuery.GetFieldNames(aList);

  finally
    lQuery.Free;
  end;  

end; {procedure, GetFieldNamesForSQL}

{------------------------------------------------------------------------------}
{ TdaNDBDataSet.GetFieldsForSQL }

procedure TdaNDBDataSet.GetFieldsForSQL(aList: TList; aSQL: TStrings);
var
  lQuery: TAsaDataset;
  lQueryField: TField;
  lField: TppField;
  liIndex: Integer;
begin
  aList.Clear;

  {create a temporary query}
  lQuery := TAsaDataset.Create(Self);
  try

    {assign databae and SQL properties}
    lQuery.Session := daGetNDBConnectionForName(DatabaseName);
    lQuery.SQL := aSQL;

    {set query to active}
    lQuery.Active := True;

    {create a TppField object for each field in the query}
    for liIndex := 0 to lQuery.FieldCount - 1 do
      begin
        lQueryField := lQuery.Fields[liIndex];

        lField := TppField.Create(nil);

        lField.FieldName := lQueryField.FieldName;
        lField.FieldAlias := lQueryField.DisplayLabel;
        lField.FieldLength := lQueryField.Size;
        lField.DataType := ppConvertFieldType(lQueryField.DataType);
        lField.DisplayWidth := lQueryField.DisplayWidth;

        aList.Add(lField);
      end;

  finally
    lQuery.Free;
  end;

end; {procedure, GetFieldsForSQL}

{******************************************************************************
 *
 ** N A T I V E D B   Q U E R Y   D A T A V I E W
 *
{******************************************************************************}

{------------------------------------------------------------------------------}
{ TdaNDBQueryDataView.Create }

constructor TdaNDBQueryDataView.Create(aOwner: TComponent);
begin

  inherited Create(aOwner);

  {notes: 1. must use ChildQuery, ChildDataSource, ChildPipeline etc.
          2. use Self as owner for Query, DataSource etc.
          3. do NOT assign a Name }

  FQuery := TdaChildNDBQuery.Create(Self);

  FDataSource := TppChildDataSource.Create(Self);
  FDataSource.DataSet := FQuery;

end; {constructor, Create}

{------------------------------------------------------------------------------}
{ TdaNDBQueryDataView.Destroy }

destructor TdaNDBQueryDataView.Destroy;
begin
  FDataSource.Free;
  FQuery.Free;

  inherited Destroy;

end; {destructor, Destroy}

{------------------------------------------------------------------------------}
{ TdaNDBQueryDataView.SessionClass }

class function TdaNDBQueryDataView.SessionClass: TClass;
begin
  Result := TdaNDBSession;
end; {class function, SessionClass}

{------------------------------------------------------------------------------}
{ TdaNDBQueryDataView.PreviewFormClass }

class function TdaNDBQueryDataView.PreviewFormClass: TFormClass;
begin
  Result := TFormClass(GetClass('TdaPreviewDataDialog'));
end; {class function, PreviewFormClass}

{------------------------------------------------------------------------------}
{ TdaIBOQueryDataView.ConnectPipelinesToData }

procedure TdaNDBQueryDataView.ConnectPipelinesToData;
begin

  if DataPipelineCount = 0 then Exit;

  {need to reconnect here}
  TppDBPipeline(DataPipelines[0]).DataSource := FDataSource;


end; {procedure, ConnectPipelinesToData}

{------------------------------------------------------------------------------}
{ TdaNDBQueryDataView.Init }

procedure TdaNDBQueryDataView.Init;
var
  lDataPipeline: TppChildDBPipeline;

begin

  inherited Init;

  if DataPipelineCount > 0 then Exit;

  {note: DataView's owner must own the DataPipeline }
  lDataPipeline := TppChildDBPipeline(ppComponentCreate(Self, TppChildDBPipeline));
  lDataPipeline.DataSource := FDataSource;
 
  lDataPipeline.AutoCreateFields := False;

  {add DataPipeline to the dataview }
  lDataPipeline.DataView := Self;

end; {procedure, Init}

{------------------------------------------------------------------------------}
{ TdaNDBQueryDataView.SQLChanged }

procedure TdaNDBQueryDataView.SQLChanged;
begin

  if FQuery.Active then
    FQuery.Close;

  FQuery.Session := daGetNDBConnectionForName(SQL.DatabaseName);
  FQuery.SQL := SQL.MagicSQLText;

end; {procedure, SQLChanged}

{******************************************************************************
 *
 ** P R O C E D U R E S   A N D   F U N C T I O N S
 *
{******************************************************************************}

{------------------------------------------------------------------------------}
{ daGetDefaultNDBConnection }

function daGetDefaultNDBConnection: TAsaSession;
begin

  {create the default Connection, if needed}
  if (FNDBConnection = nil) then
    begin
      {create default NDB Connection}
      FNDBConnection := TAsaSession.Create(nil);
      FNDBConnection.Name := cDefaultConnection;

      FNDBLogin := TAsaLogin.Create(nil);
      FNDBLogin.Session := FNDBConnection;
    end;

  Result := FNDBConnection;

end; {function, daGetDefaultNDBConnection}

{------------------------------------------------------------------------------}
{ daGetNDBConnectionNames }

procedure daGetNDBConnectionNames(aList: TStrings);
begin

  {call utility routine in NdbApp unit}
  NdbGetDatabaseNames(aList);

end; {procedure, daGetNDBConnectionNames}

{------------------------------------------------------------------------------}
{ daGetNDBConnectionForName }

function daGetNDBConnectionForName(aDatabaseName: String): TAsaSession;
var
  liIndex: Integer;

begin
  Result := nil;

  liIndex := 0;

 {check for a database object with this name}
  while (Result = nil) and (liIndex < daGetNDBConnectionList.Count) do
    begin
      if (AnsiCompareStr(FNDBConnectionList[liIndex].Name, aDatabaseName) = 0) or
         (AnsiCompareStr(TAsaSession(FNDBConnectionList[liIndex]).LoginDatabase, aDatabaseName) = 0) then
        Result :=  TAsaSession(FNDBConnectionList[liIndex]);
      Inc(liIndex);
    end;

  if (Result <> nil) then Exit;

  {use the default database object}
  Result := daGetDefaultNDBConnection;

  {set DatabaseName property, if needed}
  if (Result.LoginDatabase <> aDatabaseName) or (aDatabaseName = '') then
    begin
      if Result.Connected then
        Result.Connected := False;
      Result.LoginDatabase := aDatabaseName;

      {NDB does not auto logon, so force a logon dialog here}
      daNDBConnectToDatabase(Result);
    end;

end; {function, daGetNDBConnectionForName}

{------------------------------------------------------------------------------}
{ daNDBConnectToDatabase }

function daNDBConnectToDatabase(aConnection: TAsaSession): Boolean;
begin
  Result := False;

  if (aConnection = nil) then Exit;

  if not aConnection.Connected then
    begin
      if (aConnection = daGetDefaultNDBConnection) then
        FNDBLogin.Execute
      else
        aConnection.Connected := True;

    end;

  Result := aConnection.Connected;

end; {function, daNDBConnectToDatabase}

{------------------------------------------------------------------------------}
{ daGetNDBConnectionList }

function daGetNDBConnectionList: TppComponentList;
begin
  if (FNDBConnectionList = nil) then
    FNDBConnectionList := TppComponentList.Create(nil);

  Result := FNDBConnectionList;

end; {function, daGetNDBConnectionList}

initialization

  {register the NDB descendant classes}
  RegisterClasses([TdaChildNDBQuery, TdaChildNDBTable, TdaChildNDBStoredProc]);

  {register DADE descendant session, dataset, dataview}
  daRegisterSession(TdaNDBSession);
  daRegisterDataSet(TdaNDBDataSet);
  daRegisterDataView(TdaNDBQueryDataView);

  {initialize internal reference variables}
  FNDBConnection     := nil;
  FNDBLogin          := nil;
  FNDBConnectionList := nil;

finalization

  {free the default connection object}
  FNDBConnection.Free;
  FNDBLogin.Free;
  FNDBConnectionList.Free;

  {unregister the NDB descendant classes}
  UnRegisterClasses([TdaChildNDBQuery, TdaChildNDBTable, TdaChildNDBStoredProc]);

  {unregister DADE descendant the session, dataset, dataview}
  daUnRegisterSession(TdaNDBSession);
  daUnRegisterDataSet(TdaNDBDataSet);
  daUnRegisterDataView(TdaNDBQueryDataView);

end.
